/********************************************************************************
 * Copyright (c) 2023 CEA-List
 *
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License 2.0 which is available at
 * http://www.eclipse.org/legal/epl-2.0.
 *
 * SPDX-License-Identifier: EPL-2.0
 *
 ********************************************************************************/

#include <cassert>
#include <cstddef>
#include <cstdint>
#include <string>
#include <utility>
#include <vector>

#include "aidge/backend/OperatorImpl.hpp"
#include "aidge/operator/Slice.hpp"
#include "aidge/utils/ErrorHandling.hpp"
#include "aidge/utils/Types.h"

const std::string Aidge::Slice_Op::Type = "Slice";

void Aidge::Slice_Op::computeOutputDims() {
    // check input have been associated
    if (!getInput(0) || !getInput(1) || !getInput(2) || !getInput(3)) {
        AIDGE_THROW_OR_ABORT(std::runtime_error, "Every input should be associated with a Tensor");
    }

    if((!getInput(0)->empty()) && (!getInput(1)->empty()) && (!getInput(2)->empty()) && (!getInput(3)->empty()))
    {
        const auto starts = mInputs[1]->getImpl()->rawPtr();
        const auto ends = mInputs[2]->getImpl()->rawPtr();
        const auto axes = mInputs[3]->getImpl()->rawPtr();
        DimSize_t nbAxes = mInputs[1]->size();
        std::vector<DimSize_t> outDims = getInput(0)->dims();
        for (std::size_t i = 0; i < nbAxes; ++i) {
            // For each slice operation get the params and cast them to size_t
            std::size_t axis, start, end; //TODO find a better way to cast "starts", "ends" and "axes"
            if (mInputs[1]->dataType() == DataType::Float32 && mInputs[2]->dataType() == DataType::Float32 && mInputs[3]->dataType() == DataType::Float32)
            {
                const float* axes_ = static_cast<float*>(axes);
                axis = axes_[i] >= 0 ? static_cast<std::size_t>(axes_[i]) : static_cast<std::size_t>(axes_[i]) + getInput(0)->nbDims();
                const float* starts_ = static_cast<float*>(starts);
                start = starts_[i] >= 0 ? static_cast<std::size_t>(starts_[i]) : static_cast<std::size_t>(starts_[i]) + getInput(0)->dims()[axis];
                const float* ends_ = static_cast<float*>(ends);
                end = ends_[i] >= 0 ? static_cast<std::size_t>(ends_[i]) : static_cast<std::size_t>(ends_[i]) + getInput(0)->dims()[ends_[i]];
            }
            else if(mInputs[1]->dataType() == DataType::Int32 && mInputs[2]->dataType() == DataType::Int32 && mInputs[3]->dataType() == DataType::Int32)
            {
                const std::int32_t* axes_ = static_cast<std::int32_t*>(axes);
                axis = axes_[i] >= 0 ? static_cast<std::size_t>(axes_[i]) : static_cast<std::size_t>(axes_[i]) + getInput(0)->nbDims();
                const std::int32_t* starts_ = static_cast<std::int32_t*>(starts);
                start = starts_[i] >= 0 ? static_cast<std::size_t>(starts_[i]) : static_cast<std::size_t>(starts_[i]) + getInput(0)->dims()[axis];
                const std::int32_t* ends_ = static_cast<std::int32_t*>(ends);
                end = ends_[i] >= 0 ? static_cast<std::size_t>(ends_[i]) : static_cast<std::size_t>(ends_[i]) + getInput(0)->dims()[ends_[i]];
            }
            else
            {
                AIDGE_THROW_OR_ABORT(std::runtime_error, "Slice inputs type is not supported yet");
            }
            const std::size_t sliceLength = end - start;
            // Check if slice length is valid
            if (sliceLength > getInput(0)->dims()[axis])
            {
                AIDGE_THROW_OR_ABORT(std::runtime_error, "ROI of Slice operator out of bounds");
            }
            outDims[axis] = sliceLength;
        }
        mOutputs[0]->resize(outDims);
    }
}